import jax
import jax.numpy as jnp
from jaxtyping import Array, Float


def get_t_samples(
    key: int, num_samples: int, law: str = "uniform", **kwargs
) -> Float[Array, "num_samples 1"]:
    """Sample the t values for a given law

    Args:
        key: Random key
        num_samples: Number of samples
        law: Law to sample from
        **kwargs: Optional arguments for the law

    Raises:
        ValueError: If the law is unknown

    Returns:
        Float[Array, "num_samples 1"]: The sampled t values
    """
    if law == "normal":
        t = 0.3 * jax.random.normal(key, (num_samples, 1)) + 0.5
        # rescale to [0, 1]
        t = jnp.clip(t, 0, 1)
    elif law == "uniform":
        t = jax.random.uniform(key, (num_samples, 1))
    elif law == "power":
        alpha = kwargs.get("alpha", 0.2)
        a = (1 + alpha) * -1 + 1
        t = jax.random.beta(key, a, 1, (num_samples, 1))
    else:
        raise ValueError(f"Unknown law {law}")

    return t



import numpy as np

# Create a function that generates 8 gaussian
def eight_gaussian(num_samples: int, noise: float = 0.05) -> np.ndarray:
    """Generate 8 gaussians data

    Args:
        num_samples: Number of samples to generate
        noise: Noise level

    Returns:
        Data
    """
    centers = [
        (1, 0),
        (-1, 0),
        (0, 1),
        (0, -1),
        (1. / np.sqrt(2), 1. / np.sqrt(2)),
        (1. / np.sqrt(2), -1. / np.sqrt(2)),
        (-1. / np.sqrt(2), 1. / np.sqrt(2)),
        (-1. / np.sqrt(2), -1. / np.sqrt(2))
    ]

    dataset = []
    for i in range(num_samples):
        point = np.random.randn(2) * 0.05
        shift = np.array([0.5, 0.5])
        scale = 2
        idx = np.random.choice(len(centers))
        center = centers[idx]
        point[0] += center[0]
        point[1] += center[1]
        point = point * scale + shift
        dataset.append(point)
    dataset = np.array(dataset, dtype=np.float32)
    perm = np.random.permutation(len(dataset))
    return dataset[perm], None



"""Training utilities for rectified flows."""

from typing import List

import equinox as eqx
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import optax
from jaxtyping import Array, Float
from tqdm import tqdm

from jaxfm.flows.base import Flow


def get_loss_fn(name="l2"):
    if name == "l2":

        @eqx.filter_jit
        def loss_fn(model, z_t, t, target):
            vmapped_model = jax.vmap(model, in_axes=(0, 0))
            return jnp.sum((vmapped_model(z_t, t) - target) ** 2, axis=1).mean()

    return loss_fn


def train_flow(
    flow: Flow,
    optimizer: optax.GradientTransformation,
    x0: Float[Array, "batch_size dim"],
    x1: Float[Array, "batch_size dim"],
    batch_size: int,
    inner_iters: int,
) -> (Flow, List[float]):
    """Train the flow with data points x0 and x1.

    Args:
        flow: Instance of the flow model.
        optimizer: optimizer for training.
        x0: Source points.
        x1: Target points.
        batch_size: Batch size for training.
        inner_iters: Number of training iterations.

    Returns:
        Trained flow and loss curve.
    """
    loss_curve = []
    opt_state = optimizer.init(eqx.filter(flow.model, eqx.is_array))
    key = jax.random.PRNGKey(0)
    loss_fn = get_loss_fn()

    @eqx.filter_jit
    def make_step(model, z_t, t, target, opt_state):
        loss_value, grads = eqx.filter_value_and_grad(loss_fn)(model, z_t, t, target)
        updates, opt_state = optimizer.update(
            grads, opt_state, eqx.filter(model, eqx.is_array)
        )
        model = eqx.apply_updates(model, updates)
        return model, opt_state, loss_value

    trange = tqdm(range(inner_iters))
    for i in trange:
        key, subkey = jax.random.split(key)
        idx = jax.random.choice(subkey, len(x0), (batch_size,), replace=False)
        key, subkey = jax.random.split(key)
        z0 = x0[idx]
        z1 = x1[idx]

        z_t, t, target = flow.get_train_tuple(z0, z1, subkey)
        # print(z_t.shape, t.shape, target.shape)
        flow.model, opt_state, loss = make_step(
            flow.model, z_t, t, target, opt_state
        )
        if i == inner_iters - 1:
            vt = jax.vmap(flow.model, in_axes=(0, 0))(z_t, t)
            fig = plt.figure(figsize=(8, 5))
            angle = jnp.arctan2(vt[:, 1], vt[:, 0])
            plt.quiver(z_t[:, 0], z_t[:, 1], jnp.cos(angle), jnp.sin(angle))
            plt.savefig("fig/vt.pdf")
            plt.close(fig)
        loss_curve.append(loss)

        trange.set_postfix(loss=loss)
    return flow, loss_curve



import numpy as np

def two_moons(num_samples: int, noise: float = 0.05) -> np.ndarray:
    """Generate two moons data

    Args:
        num_samples: Number of samples to generate
        noise: Noise level

    Returns:
        Data
    """
    from sklearn.datasets import make_moons

    x, labels = make_moons(n_samples=num_samples, noise=noise)
    perm = np.random.permutation(len(x))
    return x[perm], labels[perm]



from typing import Any

from jaxfm.flows.base import Flow
from jaxfm.flows.conditional_flows import CFM
from jaxfm.flows.reflows import ReFlow


def get_flow(name: str, model: Any, num_steps: int, **kwargs) -> Flow:
    """Get the flow model."""
    if name == "reflow":
        return ReFlow(model=model, num_steps=num_steps)
    elif name == "conditional":
        return CFM(model=model, num_steps=num_steps, **kwargs)
    else:
        raise ValueError(f"Unknown flow model {name}")



from typing import Tuple

from jaxtyping import Array, Float
from jaxfm.flows.base import Flow

from jaxfm.utils.time_scheduler import get_t_samples


class ReFlow(Flow): 
    """Implement the ReFlow model.
    Reference:
    Flow Straight and Fast: Learning to Generate and Transfer Data with Rectified Flow http://arxiv.org/abs/2209.03003
    """

    def __init__(self, model=None, num_steps=100, law: str = "uniform") -> None:
        super().__init__(model, num_steps, law)

    def get_train_tuple( # type: ignore
        self,
        z0: Float[Array, "batch_size dim"],
        z1: Float[Array, "batch_size dim"],
        key: int,
    ) -> Tuple[
        Float[Array, "batch_size dim"],
        Float[Array, "batch_size 1"],
        Float[Array, "batch_size dim"],
    ]:
        """Interpolate the samples and get the target.

        Args:
            z0: Source samples
            z1: Target samples
            key: Random key

        Returns:
            Interpolated samples, time samples, target
        """
        t = get_t_samples(key, z0.shape[0], law=self.law)
        z_t = z0 * (1 - t) + z1 * t
        target = z1 - z0

        return z_t, t, target
